#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import atexit, sys

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.agx import AGX
from m1n1.agx.render import *

from m1n1 import asm

from m1n1.gpiola import GPIOLogicAnalyzer

analyzer_cpu = 1

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")
p.smp_start_secondaries()
p.mmu_init_secondary(analyzer_cpu)
iface.dev.timeout = 42

## heater code
if True:
    code = u.malloc(0x1000)

    util = asm.ARMAsm("""
    bench:
        mrs x1, CNTPCT_EL0
    1:
        sub x0, x0, #1
        cbnz x0, 1b

        mrs x2, CNTPCT_EL0
        sub x0, x2, x1
        ret
    """, code)
    iface.writemem(code, util.data)
    p.dc_cvau(code, len(util.data))
    p.ic_ivau(code, len(util.data))

    LOOPS = 80000000000
    for idx in range(2, 8):
        print(f"bench {idx}")
        p.smp_call(idx, util.bench, LOOPS)

agx = AGX(u)

mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
agx.mon = mon

sgx = agx.sgx_dev
#mon.add(sgx.gpu_region_base, sgx.gpu_region_size, "contexts")
#mon.add(sgx.gfx_shared_region_base, sgx.gfx_shared_region_size, "gfx-shared")
#mon.add(sgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

#mon.add(agx.initdasgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

atexit.register(p.reboot)
agx.start()

print("==========================================")
print("## After init")
print("==========================================")
mon.poll()
agx.poll_objects()

ctx = GPUContext(agx)
ctx.bind(63)
ctx0 = GPUContext(agx)
ctx0.bind(62)

f = GPUFrame(ctx, sys.argv[1], track=False)
f2 = GPUFrame(ctx0, sys.argv[1], track=False)

RENDERERS = 4
FRAMES = 8

renderers = []

fault_cmdbuf = f.cmdbuf.clone()
#fault_cmdbuf.depth_buffer = 0xdeadb000

for i in range(RENDERERS):
    c = ctx0 if i == 0 else ctx
    r = GPURenderer(c, 32, bm_slot=0x10 + i, queue=1)
    renderers.append(r)

    for q in (r.wq_3d, r.wq_ta):
        q.info.set_prio(2)
        q.info.push()

#for r in renderers[2:4]:
    #for q in (r.wq_3d, r.wq_ta):
        #q.info.set_prio(3)
        #q.info.push()

#for r in renderers[4:6]:
    #for q in (r.wq_3d, r.wq_ta):
        #q.info.set_prio(0)
        #q.info.push()

#for r in renderers[6:8]:
    #for q in (r.wq_3d, r.wq_ta):
        #q.info.set_prio(1)
        #q.info.push()

print("==========================================")
print("## Submitting")
print("==========================================")

for i, r in enumerate(renderers):
    for j in range(FRAMES):
        if (i, j) in ((1, 0), (2, 1), (3, 1)):
            r.submit(fault_cmdbuf)
        elif i == 0:
            r.submit(f2.cmdbuf)
        else:
            r.submit(f.cmdbuf)

print("==========================================")
print("## Submitted")
print("==========================================")

def t(addr):
    paddr = agx.uat.iotranslate(0, addr, 4)[0][0]
    if paddr is None:
        raise Exception(f"Failed to iotranslate {addr:#x}")
    return paddr

regs = {
    "ta_cmds":  t(agx.initdata.regionB.stats_ta.addrof("total_cmds")),
    "ta_ts":    t(agx.initdata.regionB.stats_ta.stats.addrof("unk_timestamp")),
}

pend_base = agx.initdata.regionC.addrof("pending_stamps")
for i in range(5):
    regs[f"st{i}_info"] = t(pend_base + i*8)
    regs[f"st{i}_val"] = t(pend_base + i*8 + 4)

for i in range(4):
    regs[f"ta{i}_cq"] = t(agx.initdata.regionB.stats_ta.stats.queues[i].addrof("cur_cmdqueue"))

regs.update({
    "pwr_status": t(agx.initdata.regionB.hwdata_a.addrof("pwr_status")),
    "pstate": t(agx.initdata.regionB.hwdata_a.addrof("cur_pstate")),
    "temp_c": t(agx.initdata.regionB.hwdata_a.addrof("temp_c")),
    "pwr_mw": t(agx.initdata.regionB.hwdata_a.addrof("avg_power_mw")),
    "pwr_ts": t(agx.initdata.regionB.hwdata_a.addrof("update_ts")),

    #"unk_10": t(agx.initdata.regionB.hwdata_a.addrof("unk_10")),
    #"unk_14": t(agx.initdata.regionB.hwdata_a.addrof("unk_14")),
    "actual_pstate": t(agx.initdata.regionB.hwdata_a.addrof("actual_pstate")),
    "tgt_pstate": t(agx.initdata.regionB.hwdata_a.addrof("tgt_pstate")),
    #"unk_40": t(agx.initdata.regionB.hwdata_a.addrof("unk_40")),
    #"unk_44": t(agx.initdata.regionB.hwdata_a.addrof("unk_44")),
    #"unk_48": t(agx.initdata.regionB.hwdata_a.addrof("unk_48")),
    "freq_mhz": t(agx.initdata.regionB.hwdata_a.addrof("freq_mhz")),

    #"unk_748.0": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")),
    #"unk_748.1": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+4),
    #"unk_748.2": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+8),
    #"unk_748.3": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+12),
    #"use_percent": t(agx.initdata.regionB.hwdata_a.addrof("use_percent")),
    #"unk_83c": t(agx.initdata.regionB.hwdata_a.addrof("unk_83c")),
    "freq_with_off": t(agx.initdata.regionB.hwdata_a.addrof("freq_with_off")),
    #"unk_ba0": t(agx.initdata.regionB.hwdata_a.addrof("unk_ba0")),
    #"unk_bb0": t(agx.initdata.regionB.hwdata_a.addrof("unk_bb0")),
    #"unk_c44": t(agx.initdata.regionB.hwdata_a.addrof("unk_c44")),
    #"unk_c58": t(agx.initdata.regionB.hwdata_a.addrof("unk_c58")),

    #"unk_3ca0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3ca0")),
    #"unk_3ca8": t(agx.initdata.regionB.hwdata_a.addrof("unk_3ca8")),
    #"unk_3cb0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3cb0")),
    #"ts_last_idle": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_idle")),
    #"ts_last_poweron": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_poweron")),
    #"ts_last_poweroff": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_poweroff")),
    #"unk_3cd0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3cd0")),

    "halt_count":  t(agx.initdata.fw_status.addrof("halt_count")),
    "halted":  t(agx.initdata.fw_status.addrof("halted")),
    "resume":  t(agx.initdata.fw_status.addrof("resume")),
    "unk_40":  t(agx.initdata.fw_status.addrof("unk_40")),
    "unk_ctr":  t(agx.initdata.fw_status.addrof("unk_ctr")),
    "unk_60":  t(agx.initdata.fw_status.addrof("unk_60")),
    "unk_70":  t(agx.initdata.fw_status.addrof("unk_70")),
    "c_118c0":  t(agx.initdata.regionC._addr + 0x118c0),
    "c_118c4":  t(agx.initdata.regionC._addr + 0x118c4),
    "c_118c8":  t(agx.initdata.regionC._addr + 0x118c8),
    "c_118cc":  t(agx.initdata.regionC._addr + 0x118cc),
    "c_118d0":  t(agx.initdata.regionC._addr + 0x118d0),
    "c_118d4":  t(agx.initdata.regionC._addr + 0x118d4),
    "c_118d8":  t(agx.initdata.regionC._addr + 0x118d8),
    "c_118dc":  t(agx.initdata.regionC._addr + 0x118dc),
    "3d_cmds":  t(agx.initdata.regionB.stats_3d.addrof("total_cmds")),
    #"3d_tvb_oflws_1":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_1")),
    #"3d_tvb_oflws_2":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_2")),
    "3d_cur_stamp_id":  t(agx.initdata.regionB.stats_3d.stats.addrof("cur_stamp_id")),
    "3d_ts":    t(agx.initdata.regionB.stats_3d.stats.addrof("unk_timestamp")),
    #"3d_cur_stamp_id":  t(agx.initdata.regionB.stats_3d.stats.addrof("cur_stamp_id")),
})

for i in range(4):
    regs[f"3d{i}_cq"] = t(agx.initdata.regionB.stats_3d.stats.queues[i].addrof("cur_cmdqueue"))


for i, r in enumerate(renderers):
    regs.update({
        f"r{i}_3d_done":  t(r.wq_3d.info.pointers.addrof("gpu_doneptr")),
        #f"r{i}_3d_rptr":  t(r.wq_3d.info.pointers.addrof("gpu_rptr")),
        f"r{i}_3d_busy":  t(r.wq_3d.info.addrof("busy")),
        #f"r{i}_3d_blk":   t(r.wq_3d.info.addrof("blocked_on_barrier")),
        #f"r{i}_3d_2c":    t(r.wq_3d.info.addrof("unk_2c")),
        #f"r{i}_3d_54":    t(r.wq_3d.info.addrof("unk_54")),

        f"r{i}_ta_done":  t(r.wq_ta.info.pointers.addrof("gpu_doneptr")),
        #f"r{i}_ta_rptr":  t(r.wq_ta.info.pointers.addrof("gpu_rptr")),
        f"r{i}_ta_busy":  t(r.wq_ta.info.addrof("busy")),
        #f"r{i}_ta_blk":   t(r.wq_ta.info.addrof("blocked_on_barrier")),
        #f"r{i}_ta_2c":    t(r.wq_ta.info.addrof("unk_2c")),
        #f"r{i}_ta_54":    t(r.wq_ta.info.addrof("unk_54")),
        f"r{i}_f{j}_ta_stamp1": t(r.stamp_ta1._addr),
        f"r{i}_ta_stamp2":t(r.stamp_ta2._addr),
        f"r{i}_f{j}_3d_stamp1": t(r.stamp_3d1._addr),
        f"r{i}_3d_stamp2":t(r.stamp_3d2._addr),
    })

    for j in range(FRAMES):
        work = r.work[j]
        regs.update({
            f"r{i}_f{j}_3d_ts": t(work.wc_3d.ts1._addr),
            f"r{i}_f{j}_ta_ts": t(work.wc_ta.ts1._addr),
        })

div=4
ticks = 24000000 // div * 25

la = GPIOLogicAnalyzer(u, regs=regs, cpu=analyzer_cpu, div=div)


print("==========================================")
print("## Poll prior to job start")
print("==========================================")

#mon.poll()
#agx.poll_objects()


print("==========================================")
print("## After start")
print("==========================================")
#agx.poll_objects()

#mon.poll()
print("==========================================")
print("## Waiting")
print("==========================================")

print("Queues:")
for i, r in enumerate(renderers):
    print(f" Renderer {i}")
    print(f"  TA: {r.wq_ta.info._addr:#x} (stamp {r.work[0].ev_ta.id})")
    #print(r.wq_ta.info)
    print(f"  3D: {r.wq_3d.info._addr:#x} (stamp {r.work[0].ev_3d.id})")
    #print(r.wq_3d.info)

print("==========================================")
print("## Run")
print("==========================================")

la.start(ticks, bufsize=0x8000000)

try:
    for r in renderers[:RENDERERS]:
        r.run()

    for r in renderers[:RENDERERS]:
        while not r.ev_3d.fired:
            agx.asc.work()
            agx.poll_channels()
            print("==========================================")
            agx.poll_objects()
            mon.poll()

        r.wait()

    #agx.poll_objects()

    #print("==========================================")
    #print("## Stop ASC")
    #print("==========================================")

    #agx.asc.stop()

    ##time.sleep(0.1)

    ##agx.poll_objects()

    #print("==========================================")
    #print("## Start ASC")
    #print("==========================================")

    #agx.asc.start()

    ##agx.poll_objects()

    #print("==========================================")
    #print("## Run 2")
    #print("==========================================")

    #for r in renderers[RENDERERS//2:]:
        #r.run()

    #for r in renderers[RENDERERS//2:]:
        #while not r.ev_3d.fired:
            #agx.asc.work()
            #agx.poll_channels()
            #print("==========================================")

        #r.wait()

    #agx.poll_objects()

    #mon.poll()

finally:
    #agx.poll_objects()
    #mon.poll()

    la.complete()
    la.show()

time.sleep(2)




#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import errno, ctypes, sys, atexit, os, os.path, mmap
from construct import *

from m1n1 import malloc
from m1n1.utils import Register32
from m1n1.agx import AGX
from m1n1.agx.render import *
from m1n1.agx.uapi import *
from m1n1.proxyutils import *
from m1n1.utils import *

PAGE_SIZE = 32768
SHIM_MEM_SIZE = 4 * 1024 * 1024 * 1024

class IOCTL(Register32):
    NR = 7, 0
    TYPE = 15, 8
    SIZE = 29, 16
    DIR = 31, 30

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2

_IO = lambda type, nr: IOCTL(TYPE=type, NR=nr, SIZE=0, DIR=_IOC_NONE)
_IOR = lambda type, nr, size: IOCTL(TYPE=type, NR=nr, SIZE=size, DIR=_IOC_READ)
_IOW = lambda type, nr, size: IOCTL(TYPE=type, NR=nr, SIZE=size, DIR=_IOC_WRITE)
_IOWR = lambda type, nr, size: IOCTL(TYPE=type, NR=nr, SIZE=size, DIR=_IOC_READ|_IOC_WRITE)

DRM_IOCTL_BASE = ord('d')

def IO(nr):
    def dec(f):
        f._ioctl = _IO(DRM_IOCTL_BASE, nr)
        return f
    return dec

def IOR(nr, cls):
    def dec(f):
        f._ioctl = _IOR(DRM_IOCTL_BASE, nr, cls.sizeof())
        f._arg_cls = cls
        return f
    return dec

def IOW(nr, cls):
    def dec(f):
        f._ioctl = _IOW(DRM_IOCTL_BASE, nr, cls.sizeof())
        f._arg_cls = cls
        return f
    return dec

def IOWR(nr, cls):
    def dec(f):
        f._ioctl = _IOWR(DRM_IOCTL_BASE, nr, cls.sizeof())
        f._arg_cls = cls
        return f
    return dec

class DRMAsahiShim:
    def __init__(self, memfd):
        self.memfd = memfd
        self.initialized = False
        self.ioctl_map = {}
        for key in dir(self):
            f = getattr(self, key)
            ioctl = getattr(f, "_ioctl", None)
            if ioctl is not None:
                self.ioctl_map[ioctl.value] = ioctl, f
        self.bos = {}
        self.pull_buffers = bool(os.getenv("ASAHI_SHIM_PULL"))
        self.dump_frames = bool(os.getenv("ASAHI_SHIM_DUMP"))
        self.frame = 0
        self.agx = None

    def read_buf(self, ptr, size):
        return ctypes.cast(ptr, ctypes.POINTER(ctypes.c_ubyte * size))[0]

    def init_agx(self):
        from m1n1.setup import p, u, iface

        p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
        p.pmgr_adt_clocks_enable("/arm-io/sgx")

        self.agx = agx = AGX(u)

        mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
        agx.mon = mon

        sgx = agx.sgx_dev
        #mon.add(sgx.gpu_region_base, sgx.gpu_region_size, "contexts")
        #mon.add(sgx.gfx_shared_region_base, sgx.gfx_shared_region_size, "gfx-shared")
        #mon.add(sgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

        #mon.add(agx.initdasgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

        atexit.register(p.reboot)
        agx.start()

    def init(self):
        if self.initialized:
            return

        self.init_agx()
        self.ctx = GPUContext(self.agx)
        self.ctx.bind(0x17)
        self.renderer = GPURenderer(self.ctx, 0x40, bm_slot=10, queue=1)

        self.initialized = True

    @IOW(DRM_COMMAND_BASE + 0x00, drm_asahi_submit_t)
    def submit(self, fd, args):
        sys.stdout.write(".")
        sys.stdout.flush()

        size = drm_asahi_cmdbuf_t.sizeof()
        cmdbuf = drm_asahi_cmdbuf_t.parse(self.read_buf(args.cmdbuf, size))

        self.log("Pushing objects...")
        for obj in self.bos.values():
            #if obj._skipped_pushes > 64:# and obj._addr > 0x1200000000 and obj._size > 131072:
                #continue
            obj.push(True)
        self.log("Push done")

        attachment_objs = []
        for i in cmdbuf.attachments:
            for obj in self.bos.values():
                if obj._addr == i.pointer:
                    attachment_objs.append(obj)

        if self.dump_frames:
            name = f"shim_frame{self.frame:03d}.agx"
            f = GPUFrame(self.renderer.ctx)
            f.cmdbuf = cmdbuf
            for obj in self.bos.values():
                f.add_object(obj)
            f.save(name)

        self.renderer.submit(cmdbuf)
        self.renderer.run()
        self.renderer.wait()

        if self.pull_buffers:
            self.log("Pulling buffers...")
            for obj in attachment_objs:
                obj.pull()
                obj._map[:] = obj.val
                obj.val = obj._map
            self.log("Pull done")

        #print("HEAP STATS")
        #self.ctx.uobj.va.check()
        #self.ctx.gobj.va.check()
        #self.ctx.pobj.va.check()
        #self.agx.kobj.va.check()
        #self.agx.cmdbuf.va.check()
        #self.agx.kshared.va.check()
        #self.agx.kshared2.va.check()

        self.frame += 1
        return 0

    @IOW(DRM_COMMAND_BASE + 0x01, drm_asahi_wait_bo_t)
    def wait_bo(self, fd, args):
        self.log("Wait BO!", args)
        return 0

    @IOWR(DRM_COMMAND_BASE + 0x02, drm_asahi_create_bo_t)
    def create_bo(self, fd, args):
        memfd_offset = args.offset

        if args.flags & ASAHI_BO_PIPELINE:
            alloc = self.renderer.ctx.pobj
        else:
            alloc = self.renderer.ctx.gobj

        obj = alloc.new(args.size, name=f"GBM offset {memfd_offset:#x}", track=False)
        obj._memfd_offset = memfd_offset
        obj._pushed = False
        obj.val = obj._map = mmap.mmap(self.memfd, args.size, offset=memfd_offset)
        self.bos[memfd_offset] = obj
        args.offset = obj._addr

        if args.flags & ASAHI_BO_PIPELINE:
            args.offset -= self.renderer.ctx.pipeline_base

        self.log(f"Create BO @ {memfd_offset:#x}")
        return 0

    @IOWR(DRM_COMMAND_BASE + 0x04, drm_asahi_get_param_t)
    def get_param(self, fd, args):
        self.log("Get Param!", args)
        return 0

    @IOWR(DRM_COMMAND_BASE + 0x05, drm_asahi_get_bo_offset_t)
    def get_bo_offset(self, fd, args):
        self.log("Get BO Offset!", args)
        return 0

    def bo_free(self, memfd_offset):
        self.log(f"Free BO @ {memfd_offset:#x}")
        self.bos[memfd_offset].free()
        del self.bos[memfd_offset]
        sys.stdout.flush()

    def ioctl(self, fd, request, p_arg):
        self.init()

        p_arg = ctypes.c_void_p(p_arg)

        if request not in self.ioctl_map:
            self.log(f"Unknown ioctl: fd={fd} request={IOCTL(request)} arg={p_arg:#x}")
            return -errno.ENOSYS

        ioctl, f = self.ioctl_map[request]

        size = ioctl.SIZE
        if ioctl.DIR & _IOC_WRITE:
            args = f._arg_cls.parse(self.read_buf(p_arg, size))
            ret = f(fd, args)
        elif ioctl.DIR & _IOC_READ:
            args = f._arg_cls.parse(bytes(size))
            ret = f(fd, args)
        else:
            ret = f(fd)

        if ioctl.DIR & _IOC_READ:
            data = args.build()
            assert len(data) == size
            ctypes.memmove(p_arg, data, size)

        sys.stdout.flush()
        return ret

    def log(self, s):
        if self.agx is None:
            print("[Shim] " + s)
        else:
            self.agx.log("[Shim] " + s)

Shim = DRMAsahiShim



from enum import IntEnum
from construct import *
from io import BytesIO

from ..afk.epic import *
from m1n1.utils import FourCC, chexdump
from m1n1.constructutils import *

class AOPPropKey(IntEnum):
    IS_READY     = 0x01
    MANUFACTURER = 0x0f  # wtf is a firefish2?
    CHIP_ID      = 0x11
    PLACEMENT    = 0x1e
    UNK_21       = 0x21
    ORIENTATION  = 0x2e
    LOCATION_ID  = 0x30
    PRODUCT_ID2  = 0x3f
    SERIAL_NO    = 0x3e
    CHANNEL_NAME = 0x45
    VENDOR_ID    = 0x5a
    PRODUCT_ID   = 0x5b
    SERVICE_CONTROLLER = 0x64
    DEVICE_COUNT = 0x65
    VERSION      = 0x67
    UNK_DUMP     = 0xd7

class EPICCall:
    @classmethod
    def matches(cls, hdr, sub):
        return int(sub.type) == cls.TYPE

    def _args_fixup(self):
        pass

    def __init__(self, *args, **kwargs):
        if args:
            self.args = args[0]
        else:
            self.args = Container(**kwargs)
            self._args_fixup()
        self.rets = None

    @classmethod
    def from_stream(cls, f):
        return cls(cls.ARGS.parse_stream(f))

    def dump(self, logger=print):
        args_fmt = [f"{k}={v}" for (k, v) in self.args.items() if k != "_io"]
        rets_fmt = [f"{k}={v}" for (k, v) in self.rets.items() if k != "_io"]
        logger(f"{type(self).__name__}({', '.join(args_fmt)}) -> ({', '.join(rets_fmt)})")

    def read_resp(self, f):
        self.rets = self.RETS.parse_stream(f)

CALLTYPES = []
def reg_calltype(calltype):
    CALLTYPES.append(calltype)
    return calltype

@reg_calltype
class GetHIDDescriptor(EPICCall):
    TYPE = 0x1
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
    )
    RETS = Struct(
        "retcode" / Default(Hex(Int32ul), 0),
        "descriptor" / HexDump(GreedyBytes),
    )

@reg_calltype
class GetProperty(EPICCall):
    TYPE = 0xa
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Enum(Int32ul, AOPPropKey),
    )
    RETS = Struct(
        #"retcode" / Const(0x0, Int32ul),
        "value" / GreedyBytes,
    )

class GetPropertyIsReady(EPICCall):
    TYPE = 0xa
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Const(AOPPropKey.IS_READY, Int32ul),
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "state" / FourCC,
    )

class ALSPropertyKey(IntEnum):
    INTERVAL      = 0x00
    CALIBRATION   = 0x0b
    MODE          = 0xd7
    VERBOSITY     = 0xe1
    UNKE4         = 0xe4

@reg_calltype
class ALSSetProperty(EPICCall):
    TYPE = 0x4
    SUBCLASSES = {}

    @classmethod
    def subclass(cls, cls2):
        cls.SUBCLASSES[int(cls2.SUBTYPE)] = cls2
        return cls2

@ALSSetProperty.subclass
class ALSSetPropertyVerbosity(ALSSetProperty):
    SUBTYPE = ALSPropertyKey.VERBOSITY
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Default(Hex(Int32ul), ALSPropertyKey.VERBOSITY),
        "level" / Hex(Int32ul),
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "value" / GreedyBytes,
    )

@ALSSetProperty.subclass
class ALSSetPropertyMode(ALSSetProperty):
    SUBTYPE = ALSPropertyKey.MODE
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Default(Hex(Int32ul), ALSPropertyKey.MODE),
        "mode" / Int32ul,
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "value" / GreedyBytes,
    )

@ALSSetProperty.subclass
class ALSSetPropertyCalibration(ALSSetProperty):
    SUBTYPE = ALSPropertyKey.CALIBRATION
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Default(Hex(Int32ul), ALSPropertyKey.CALIBRATION),
        "value" / GreedyBytes,
    )
    RETS = Struct(
        "retcode" / Const(0xE00002BC, Hex(Int32ul)),
    )

@ALSSetProperty.subclass
class ALSSetPropertyInterval(ALSSetProperty):
    SUBTYPE = ALSPropertyKey.INTERVAL
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Default(Hex(Int32ul), ALSPropertyKey.INTERVAL),
        "interval" / Int32ul,
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
    )

@ALSSetProperty.subclass
class ALSSetPropertyUnkE4(ALSSetProperty):
    SUBTYPE = ALSPropertyKey.UNKE4
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "key" / Default(Hex(Int32ul), ALSPropertyKey.UNKE4),
        "value" / Default(Hex(Int32ul), 0x1),
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
    )

ALSLuxReport = Struct(
    "unk0" / Const(0xec, Hex(Int8ul)),
    "sequence" / Int32ul,
    "timestamp" / Hex(Int64ul),
    "red" / Int32ul,
    "green" / Int32ul,
    "blue" / Int32ul,
    "clear" / Int32ul,
    "lux" / Float32l,
    "unk_zero" / Int32ul, # 0
    "status" / Int32ul, # 3
    "gain" / Int16ul,
    "unk3" / Int8ul,
    "unk4" / Int8ul,
    "unk5" / Int16ul,
    "integration_time" / Int32ul,
)

@reg_calltype
class WrappedCall(EPICCall):
    SUBCLASSES = {}
    TYPE = 0x20
    HDR = Struct(
        "blank" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Int32ul),
        "blank2" / ZPadding(16),
        "pad" / Hex(Int32ul),
        "len" / Hex(Int64ul),
        "residue" / HexDump(GreedyBytes),
    )

    @classmethod
    def from_stream(cls, f):
        payload = f.read()
        subsub = cls.HDR.parse(payload)
        calltype = int(subsub.calltype)
        subcls = cls.SUBCLASSES.get(calltype, None)
        if subcls is None:
            raise ValueError(f"unknown calltype {calltype:#x}")
        return subcls(subcls.ARGS.parse(payload))

    @classmethod
    def reg_subclass(cls, cls2):
        cls.SUBCLASSES[int(cls2.CALLTYPE)] = cls2
        return cls2

    @classmethod
    def matches(cls, hdr, sub):
        return sub.category == EPICCategory.NOTIFY and sub.type == cls.TYPE

    def check_retcode(self):
        if self.rets.retcode:
            self.dump()
            raise ValueError(f"retcode {self.rets.retcode} in {str(type(self))} (call dumped, see above)")

@reg_calltype
class AudioProbeDevice(EPICCall):
    TYPE = 0x20
    SUBTYPE = 0xc3_00_00_01
    ARGS = Struct(
        "pad" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "subtype" / Hex(Const(0xc3000001, Int32ul)),
        "pad2" / ZPadding(16),
        "cookie" / Default(Hex(Int32ul), 0),
        "len" / Hex(Const(0x28, Int64ul)),
        "devno" / Int32ul,
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "devid" / FourCC,
        "format" / FourCC,
        "sample_rate" / Int32ul,
        "channels" / Int32ul,
        "bytes_per_sample" / Int32ul,
        "unk_1f" / Default(Hex(Int32ul), 0),
        "unk" / Array(51, Default(Hex(Int32ul), 0)),
    )

@WrappedCall.reg_subclass
class AttachDevice(WrappedCall):
    CALLTYPE = 0xc3_00_00_02
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000002, Int32ul)),
        "blank2" / ZPadding(16),
        "pad" / Padding(4),
        "len" / Hex(Const(0x2c, Int64ul)),
        "devid" / FourCC,
        "pad" / Padding(4),
    )
    RETS = Struct(
        "retcode" / Default(Hex(Int32ul), 0),
        "unk" / HexDump(GreedyBytes),
    )

@reg_calltype
class AudioAttachDevice(EPICCall):
    TYPE = 0x20
    SUBTYPE = 0xc3_00_00_02
    ARGS = Struct(
        "pad" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "subtype" / Hex(Const(0xc3000002, Int32ul)),
        "pad2" / ZPadding(16),
        "cookie" / Default(Hex(Int32ul), 0),
        "len" / Hex(Const(0x2c, Int64ul)),
        "devid" / FourCC,
        "dev2" / Default(Hex(Int32ul), 0),
    )
    RETS = Struct(
        "retcode" / Default(Hex(Int32ul), 0),
        "unk" / HexDump(GreedyBytes),
    )

PDMConfig = Struct(
    "bytesPerSample" / Int32ul,
    "clockSource" / FourCC,
    "pdmFrequency" / Int32ul,
    "pdmcFrequency" / Int32ul,
    "slowClockSpeed" / Int32ul,
    "fastClockSpeed" / Int32ul,
    "channelPolaritySelect" / Int32ul,
    "channelPhaseSelect" / Int32ul,
    "unk8" / Hex(Int32ul),
    "unk9" / Hex(Int16ul),
    "ratios" / Struct(
        "r1" / Int8ul,
        "r2" / Int8ul,
        "r3" / Int8ul,
        "pad" / Const(0, Int8ul),
    ),
    "filterLengths" / Hex(Int32ul),
    "coeff_bulk" / Int32ul,
    #"coefficients" / Struct(
    #    "c1" / Int32sl[this._.ratios.r3 * 4 + 4],
    #    "c2" / Int32sl[this._.ratios.r2 * 4 + 4],
    #    "c3" / Int32sl[this._.ratios.r1 * 4 + 4],
    #),
    #"junk" / Padding(
    #    this.coeff_bulk * 4 - 48 \
    #    - (this.ratios.r1 + this.ratios.r2 + this.ratios.r3) * 16
    #),
    "coefficients" / Int32sl[
        (this.ratios.r1 + this.ratios.r2 + this.ratios.r3) * 4 + 12
    ],
    "junk" / Padding(
        lambda this: max(0,
            this.coeff_bulk * 4 - 48 \
            - (this.ratios.r1 + this.ratios.r2 + this.ratios.r3) * 16
        )
    ),
    "unk10" / Int32ul, # maybe
    "micTurnOnTimeMs" / Int32ul,
    "blank" / ZPadding(16),
    "unk11" / Int32ul,
    "micSettleTimeMs" / Int32ul,
    "blank2" / ZPadding(69),
)

DecimatorConfig = Struct(
    "latency" / Int32ul,
    "ratios" / Struct(
        "r1" / Int8ul,
        "r2" / Int8ul,
        "r3" / Int8ul,
        "pad" / Default(Int8ul, 0),
    ),
    "filterLengths" / Hex(Int32ul),
    "coeff_bulk" / Int32ul,
    "coefficients" / Int32sl[
        (this.ratios.r1 + this.ratios.r2 + this.ratios.r3) * 4 + 12
    ],
    "junk" / Padding(
        lambda this: max(0,
            this.coeff_bulk * 4 - 48 \
            - (this.ratios.r1 + this.ratios.r2 + this.ratios.r3) * 16
        )
    ),
)

PowerSetting = Struct(
    "devid" / FourCC,
    "cookie" / Int32ul,
    "unk" / Default(Hex(Int32ul), 0),
    "blank" / ZPadding(8),
    "target_pstate" / FourCC,
    "unk2" / Int32ul,
    "blank2" / ZPadding(20),
)

DEVPROPS = {
    ('hpai', 202): PowerSetting,
    ('lpai', 202): PowerSetting,
    ('hpai', 200): FourCC,
    ('lpai', 200): FourCC,
    ('pdm0', 200): PDMConfig,
    ('pdm0', 210): DecimatorConfig,
    ('lpai', 301): Struct(
        "unk1" / Int32ul,
        "unk2" / Int32ul,
        "unk3" / Int32ul,
        "unk4" / Int32ul,
    ),
}

class AudioPropertyKey(IntEnum):
    STATE   = 200   # 0xc8
    POWER   = 202   # 0xca
    MAIN    = 203   # 0xcb
    FORMAT  = 302   # 0x12e

@reg_calltype
class AudioProperty(EPICCall):
    TYPE = 0x20
    SUBTYPE = 0xc3000004
    SUBCLASSES = {}

    @classmethod
    def subclass(cls, cls2):
        cls.SUBCLASSES[int(cls2.SUBTYPE)] = cls2
        return cls2

@AudioProperty.subclass
class AudioPropertyState(AudioProperty):
    SUBSUBTYPE = AudioPropertyKey.STATE
    ARGS = Struct(
        "pad" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000004, Int32ul)),
        "blank2" / ZPadding(16),
        "cookie" / Default(Hex(Int32ul), 0),
        "len" / Hex(Const(0x30, Int64ul)),
        "devid" / FourCC,
        "modifier" / Const(AudioPropertyKey.STATE, Int32ul),
        "data" / Const(0x1, Int32ul),
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "size" / Const(4, Int32ul),
        "state" / FourCC,
    )

@AudioProperty.subclass
class AudioPropertyFormat(AudioProperty):
    SUBSUBTYPE = AudioPropertyKey.FORMAT
    ARGS = Struct(
        "pad" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000004, Int32ul)),
        "blank2" / ZPadding(16),
        "cookie" / Default(Hex(Int32ul), 0),
        "len" / Hex(Const(0x30, Int64ul)),
        "devid" / FourCC,
        "modifier" / Const(AudioPropertyKey.FORMAT, Int32ul),
        "data" / Const(0x1, Int32ul),
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "format" / Int32ul, # 16 == float32?
        "fourcc" / FourCC,  # PCML
        "sample_rate" / Int32ul, # 16000
        "channels" / Int32ul, # 3
        "bytes_per_sample" / Int32ul, # 2
    )

AudioPowerSetting = Struct(
    "devid" / FourCC,
    "unk1" / Int32ul,
    "cookie" / Default(Hex(Int32ul), 0),
    "blank" / ZPadding(8),
    "target_pstate" / FourCC,
    "unk2" / Int32ul,
    "blank2" / ZPadding(20),
)

@AudioProperty.subclass
class AudioPropertyPower(AudioProperty):
    SUBSUBTYPE = AudioPropertyKey.POWER
    ARGS = Struct(
        "pad" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000005, Int32ul)),
        "blank2" / ZPadding(16),
        "cookie" / Default(Hex(Int32ul), 0),
        "len" / Hex(Const(0x30 + 0x30, Int64ul)), # len(this.data) + 0x30
        "devid" / FourCC,
        "modifier" / Const(AudioPropertyKey.POWER, Int32ul),
        "len2" / Hex(Const(0x30, Int32ul)), # len(this.data)
        "data" / AudioPowerSetting,
    )
    RETS = Struct(
        "retcode" / Const(0x0, Int32ul),
        "value" / HexDump(GreedyBytes),
    )

@WrappedCall.reg_subclass
class GetDeviceProp(WrappedCall):
    CALLTYPE = 0xc3_00_00_04
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000004, Int32ul)),
        "blank2" / ZPadding(16),
        "pad" / Padding(4),
        "len" / Hex(Const(0x30, Int64ul)),
        "devid" / FourCC,
        "modifier" / Int32ul,
        "unk6" / Hex(Const(0x01, Int32ul)),
    )
    RETS = Struct(
        "retcode" / Default(Hex(Int32ul), 0),
        "len" / Optional(Int32ul),
        "data" / Switch(lambda s: (s._params.devid, s._params.modifier),
            DEVPROPS,
        default=HexDump(GreedyBytes))
    )

    def read_resp(self, f):
        self.rets = self.RETS.parse_stream(f,
            devid=self.args.devid, modifier=self.args.modifier
        )

@WrappedCall.reg_subclass
class SetDeviceProp(WrappedCall):
    CALLTYPE = 0xc3_00_00_05
    ARGS = Struct(
        "blank" / Const(0x0, Int32ul),
        "unk1" / Hex(Const(0xffffffff, Int32ul)),
        "calltype" / Hex(Const(0xc3000005, Int32ul)),
        "blank2" / ZPadding(16),
        "pad" / Padding(4),
        "len" / Hex(Int64ul), # len(this.data) + 0x30
        "devid" / FourCC,
        "modifier" / Int32ul,
        "len2" / Hex(Int32ul), # len(this.data)
        "data" / Switch(lambda s: (s.devid, s.modifier),
            DEVPROPS,
        default=HexDump(GreedyBytes))
    )
    RETS = Struct(
        "retcode" / Default(Hex(Int32ul), 0),
        "unk" / HexDump(GreedyBytes),
    )

    def _args_fixup(self):
        data_len = len(self.ARGS.build(Container(len=0, len2=0, **self.args))) - 52
        if 'len' not in self.args:
            self.args.len = data_len + 0x30
        if 'len2' not in self.args:
            self.args.len2 = data_len

@reg_calltype
class IndirectCall(EPICCall):
    ARGS = EPICCmd
    RETS = EPICCmd

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.txbuf = None
        self.rxbuf = None

    @classmethod
    def matches(cls, hdr, sub):
        return sub.category == EPICCategory.COMMAND

    def read_txbuf(self, ep):
        cmd = self.args
        ep.dart.invalidate_cache()
        self.txbuf = ep.dart.ioread(0, cmd.txbuf, cmd.txlen)

        # dump the command data for offline replays of traces
        ep.log(f"===COMMAND TX DATA=== addr={cmd.txbuf:#x}")
        chexdump(self.txbuf)
        ep.log(f"===END DATA===")

    def read_rxbuf(self, ep):
        cmd = self.rets
        ep.dart.invalidate_cache()
        self.rxbuf = ep.dart.ioread(0, cmd.rxbuf, cmd.rxlen)

        ep.log(f"===COMMAND RX DATA=== addr={cmd.rxbuf:#x}")
        chexdump(self.rxbuf)
        ep.log(f"===END DATA===")

    def unwrap(self):
        fd = BytesIO()
        fd.write(b"\x00\x00\x00\x00")
        fd.write(self.txbuf)
        fd.seek(0)
        wrapped = WrappedCall.from_stream(fd)
        fd = BytesIO()
        fd.write(b"\x00\x00\x00\x00")
        fd.write(self.rxbuf)
        fd.seek(0)
        wrapped.read_resp(fd)
        return wrapped



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import argparse
from m1n1.hw.dart import DART
from m1n1.hw.admac import *

argparser = argparse.ArgumentParser()
argparser.add_argument("-b", "--bufsize", type=int, default=1024*32,
                       help="size of one DMA buffer (covered by one descriptor)")
argparser.add_argument("-n", "--node", type=str, default="admac-sio",
                       help="name of ADT node")
argparser.add_argument("-c", "--chan", "--channel", type=int, default=0,
                       help="channel no")
argparser.add_argument("-w", "--buswidth", type=E_BUSWIDTH, default=E_BUSWIDTH.W_32BIT,
                       help="DMA device-facing bus width")
argparser.add_argument("-v", "--verbose", action='store_true')
args = argparser.parse_args()

from m1n1.setup import p, u

# find full ADT path
path = None
for node in u.adt["/arm-io"]:
    if node.name == args.node:
        path = node._path.removeprefix("/device-tree")
        print(f"Found {path}", file=sys.stderr)
        if "clock-gates" in node._properties:
            print(f"Enabling {path}", file=sys.stderr)
            p.pmgr_adt_clocks_enable(path)
        break

if path is None:
    print(f"No instance named {args.node:r} found!", file=sys.stderr)
    sys.exit(1)

admac_node = u.adt[path]

iommu_mappers = dict()
for node in u.adt["/arm-io"].walk_tree():
    if "compatible" in node._properties and node.compatible == ["iommu-mapper"]:
        iommu_mappers[getattr(node, "AAPL,phandle")] = node

mapper = iommu_mappers[admac_node.iommu_parent]
dart_path = mapper._parent_path.removeprefix("/device-tree")[:-1]
dart_idx = mapper.reg

if "clock-gates" in u.adt[dart_path]._properties:
    print(f"Enabling {dart_path}", file=sys.stderr)
    p.pmgr_adt_clocks_enable(path)

dart = DART.from_adt(u, dart_path)
admac = ADMAC(u, admac_node.get_reg(0)[0], dart,
              dart_stream=dart_idx, debug=args.verbose)

chan = admac.chans[args.chan]
chan.disable()
chan.reset()
chan.read_reports()
chan.buswidth = args.buswidth
chan.framesize = E_FRAME.F_1_WORD
chan.sram_carveout = (0x0, 0x1000)

if chan.tx:
    chan.submit(bytearray(args.bufsize))
else:
    chan.submit(buflen=args.bufsize)
chan.enable()

try:
    if chan.tx:
        while (buf := sys.stdin.buffer.read(args.bufsize)):
            while not chan.can_submit():
                chan.poll()
            chan.submit(buf)
    else:
        while True:
            while chan.can_submit():
                chan.submit(buflen=args.bufsize)
            sys.stdout.buffer.write(chan.poll())
except KeyboardInterrupt:
    pass

chan.disable()



# SPDX-License-Identifier: MIT
import cv2
import numpy as np
import time
import os

MAX_TRIES = 100
def round_up(x, y): return ((x + (y - 1)) & (-y))
def round_down(x, y): return (x - (x % y))

class AVDFrame:
    def __init__(self, img, sl, y_data, uv_data):
        self.img = img
        self.sl = sl
        self.y_data = y_data
        self.uv_data = uv_data

class AVDDec:
    def __init__(self, avd):
        self.avd = avd
        self.frames = []
        self.last_poc = -1
        self.winname = "img"
        self.count = 0

    def log(self, x):
        return self.avd.log(x)

    def setup_dma(self, ctx, sl):
        avd_w32 = self.avd.avd_w32; avd_r32 = self.avd.avd_r32
        #self.avd.avd_dma_tunables_stage0()
        assert((ctx.inst_fifo_idx >= 0) and (ctx.inst_fifo_idx < ctx.inst_fifo_count))
        avd_w32(0x1104068 + (ctx.inst_fifo_idx * 0x4), ctx.inst_fifo_iova >> 8)
        avd_w32(0x1104084 + (ctx.inst_fifo_idx * 0x4), 0x100000)
        avd_w32(0x11040a0 + (ctx.inst_fifo_idx * 0x4), 0x0)
        avd_w32(0x11040bc + (ctx.inst_fifo_idx * 0x4), 0x0)
        if (sl.mode == "h265"):
            x = 0x7
            avd_w32(0x1104040, 0x0)
        if (sl.mode == "h264"):
            x = 0x1c00
            avd_w32(0x1104048, 0x0)
        if (sl.mode == "vp09"):
            x = 0x38000
            avd_w32(0x110404c, 0x0)
        avd_w32(0x110405c, avd_r32(0x110405c) | x)
        self.avd.poll()

    def get_nv12_disp_frame(self, ctx, sl, r=2):
        w = round_up(ctx.orig_width, 16)
        h = round_up(ctx.orig_height, 16)
        y_data = self.avd.ioread(ctx.y_addr & 0xffffff00, w * h, stream=0)
        y = np.frombuffer(y_data, dtype=np.uint8).reshape((h, w))
        uv_data = self.avd.ioread(ctx.uv_addr & 0xffffff00, w * (h // 2), stream=0)
        uv = np.frombuffer(uv_data, dtype=np.uint8).reshape((h // 2, w))
        u2 = cv2.resize(uv[:,::2], (w, h), interpolation=cv2.INTER_AREA)
        v2 = cv2.resize(uv[:,1::2], (w, h), interpolation=cv2.INTER_AREA)
        yuv = np.stack((y, u2, v2), axis=-1)
        img = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)[:ctx.orig_height, :ctx.orig_width, :]
        return AVDFrame(img, sl, y_data, uv_data)

    def get_nv12_disp_frame2(self, ctx, sl):
        fmt = ctx.fmt
        w = fmt.in_width
        h = fmt.in_height
        y_data = self.avd.ioread(ctx.y_addr & 0xffffff00, w * h, stream=0)
        y = np.frombuffer(y_data, dtype=np.uint8).reshape((h, w))
        if (fmt.chroma <= 1):
            ch = h // 2
        elif (fmt.chroma == 2):
            ch = h
        uv_data = self.avd.ioread(ctx.uv_addr & 0xffffff00, w * ch, stream=0)
        uv = np.frombuffer(uv_data, dtype=np.uint8).reshape((ch, w))
        u2 = cv2.resize(uv[:,::2], (w, h), interpolation=cv2.INTER_AREA)
        v2 = cv2.resize(uv[:,1::2], (w, h), interpolation=cv2.INTER_AREA)
        yuv = np.stack((y, u2, v2), axis=-1)
        img = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)[fmt.y0:fmt.y1, fmt.x0:fmt.x1,:]
        return AVDFrame(img, sl, y_data, uv_data)

    def set_insn(self, x):
        raise ValueError()

    def set_payload(self, ctx, sl):
        self.avd.iowrite(ctx.slice_data_addr, sl.get_payload(), stream=0)
        self.avd.iomon.poll()

    def get_disp_frame(self, ctx, sl):
        raise ValueError()

    def display(self, frame):
        cv2.imshow(self.winname, frame.img); cv2.waitKey(1)
        if (frame.sl.mode == "h264") or (frame.sl.mode == "h265"):
            self.last_poc = frame.sl.pic.poc
        else:
            self.last_poc = 0
        self.frames = [f for f in self.frames if f != frame]
        self.count += 1

    def select_disp_frame(self, ctx, sl):
        return self.frames[0]

    def decode(self, ctx, sl, inst_stream, display):
        if not inst_stream: return
        self.set_payload(ctx, sl)
        self.setup_dma(ctx, sl)
        for x in inst_stream:
            v = x if isinstance(x, int) else x.val
            self.set_insn(v)
        self.get_disp_frame(ctx, sl)
        assert(self.avd.avd_r32(0x1104060) == 0x2842108)
        if (display):
            if (hasattr(ctx, "fmt")):
                frame = self.get_nv12_disp_frame2(ctx, sl)
            else:
                frame = self.get_nv12_disp_frame(ctx, sl)
            self.frames.append(frame)
            frame = self.select_disp_frame(ctx, sl)
            if (frame != None):
                self.display(frame)
            return frame
        return None

class AVDH265Dec(AVDDec):
    def __init__(self, avd):
        super().__init__(avd)

    def set_insn(self, x):
        self.avd.avd_w32(0x1104004, x)

    def set_payload(self, ctx, sl):
        self.avd.iowrite(sl.payload_addr, sl.get_payload(), stream=0)
        for seg in sl.slices:
            self.avd.iowrite(seg.payload_addr, seg.get_payload(), stream=0)
        self.avd.iomon.poll()

    def get_disp_frame(self, ctx, sl):
        avd_w32 = self.avd.avd_w32; avd_r32 = self.avd.avd_r32
        avd_w32(0x1104014, 0x2b000100 | ctx.inst_fifo_idx * 0x10 | 7)
        if (ctx.pos > 1):
            for n in range(ctx.pos - 1):
                avd_w32(0x1104014, 0x2b000000 | ctx.inst_fifo_idx * 0x10 | 7)
        self.avd.poll(); self.avd.iomon.poll()

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0xc00000 == 0xc00000): # 0x2c4210c -> 0x2c4210c
                break
            self.log("[H265] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1): raise RuntimeError("error")
        avd_w32(0x1104060, 0x4)

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0x3000 == 0x2000): # 0x2c4210c -> 0x2c42108
                break
            self.log("[H265] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1): raise RuntimeError("error")
        avd_w32(0x1104060, 0x400000)  # 0x2c42108 -> 0x2842108
        self.avd.poll(); self.avd.iomon.poll()

    def select_disp_frame(self, ctx, sl):
        dpb_size = ctx.vps_list[0].vps_max_num_reorder_pics + 1
        if (len(self.frames) >= dpb_size):
            frames = [f for f in self.frames if f.sl.pic.poc == self.last_poc + 1]
            if (len(frames) == 1):
                    return frames[0]
            frames = sorted(self.frames, key=lambda f: (f.sl.pic.poc))
            return frames[0]
        return None

class AVDH264Dec(AVDDec):
    def __init__(self, avd):
        super().__init__(avd)

    def set_insn(self, x):
        self.avd.avd_w32(0x110400c, x)

    def set_payload(self, ctx, sl):
        self.avd.iowrite(ctx.slice_data_addr, sl.get_payload(), stream=0)
        self.avd.iomon.poll()

    def get_disp_frame(self, ctx, sl):
        avd_w32 = self.avd.avd_w32; avd_r32 = self.avd.avd_r32
        avd_w32(0x1104014, 0x2b000100 | (ctx.inst_fifo_idx * 0x10) | 7)
        self.avd.poll(); self.avd.iomon.poll()

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0xc00000 == 0xc00000): # 0x2843108 -> 0x2c43108
                break
            self.log("[H264] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1): raise RuntimeError("error")
        avd_w32(0x1104060, 0x1000)

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0x3000 == 0x2000): # 0x2c43108 -> 0x2c42108
                break
            self.log("[H264] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1): raise RuntimeError("error")
        avd_w32(0x1104060, 0x400000)  # 0x2c42108 -> 0x2842108
        self.avd.poll(); self.avd.iomon.poll()

    def select_disp_frame(self, ctx, sl):
        dpb_size = ctx.num_reorder_frames
        if (len(self.frames) >= dpb_size):
                frames = [f for f in self.frames if f.sl.pic.poc == self.last_poc + 2]
                if (len(frames) == 1):
                    return frames[0]
                    return
                frames = sorted(self.frames, key=lambda f: (f.sl.pic.poc))
                return frames[0]
        return None

class AVDVP9Dec(AVDDec):
    def __init__(self, avd):
        super().__init__(avd)

    def set_insn(self, x):
        self.avd.avd_w32(0x1104010, x)

    def set_payload(self, ctx, sl):
        self.avd.iowrite(ctx.slice_data_addr, sl.get_payload(), stream=0)
        self.avd.iowrite(ctx.probs_addr, sl.get_probs(), stream=0)
        self.avd.iomon.poll()

    def get_disp_frame(self, ctx, sl):
        avd_w32 = self.avd.avd_w32; avd_r32 = self.avd.avd_r32
        avd_w32(0x1104014, 0x2bfff100 | (ctx.inst_fifo_idx * 0x10) | 7)
        if (len(sl.tiles) > 1):
            for n in range(len(sl.tiles) - 1):
                avd_w32(0x1104014, 0x2bfff000 | (ctx.inst_fifo_idx * 0x10) | 7)
        self.avd.poll(); self.avd.iomon.poll()

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0xc00000 == 0xc00000): # 0x2862108 -> 0x2c62108
                break
            self.log("[VP9] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1):
                raise RuntimeError("error")
        avd_w32(0x1104060, 0x20000)

        for n in range(MAX_TRIES):
            status = avd_r32(0x1104060)
            if (status & 0x3000 == 0x2000): # 0x2c62108 -> 0x2c42108
                break
            self.log("[VP9] status: 0x%x" % (status))
            if (n >= MAX_TRIES - 1):
                raise RuntimeError("error")
        avd_w32(0x1104060, 0x400000)  # 0x2c42108 -> 0x2842108
        self.avd.poll(); self.avd.iomon.poll()



# SPDX-License-Identifier: MIT
from enum import Enum
import threading, traceback, bisect, copy, heapq, importlib, sys, itertools, time, os, functools, struct, re, signal
from construct import Adapter, Int64ul, Int32ul, Int16ul, Int8ul, ExprAdapter, GreedyRange, ListContainer, StopFieldError, ExplicitError, StreamError

__all__ = ["FourCC"]

def align_up(v, a=16384):
    return (v + a - 1) & ~(a - 1)

align = align_up

def align_down(v, a=16384):
    return v & ~(a - 1)

def align_pot(v):
    out = 1
    while out < v:
        out *= 2
    return out

def hexdump(s, sep=" "):
    return sep.join(["%02x"%x for x in s])

def hexdump32(s, sep=" "):
    vals = struct.unpack("<%dI" % (len(s)//4), s)
    return sep.join(["%08x"%x for x in vals])

def _ascii(s):
    s2 = ""
    for c in s:
        if c < 0x20 or c > 0x7e:
            s2 += "."
        else:
            s2 += chr(c)
    return s2

def chexdump(s, st=0, abbreviate=True, stride=16, indent="", print_fn=print):
    last = None
    skip = False
    for i in range(0,len(s),stride):
        val = s[i:i+stride]
        if val == last and abbreviate:
            if not skip:
                print_fn(indent+"%08x  *" % (i + st))
                skip = True
        else:
            print_fn(indent+"%08x  %s  |%s|" % (
                i + st,
                "  ".join(hexdump(val[i:i+8], ' ').ljust(23)
                          for i in range(0, stride, 8)),
                _ascii(val).ljust(stride)))
            last = val
            skip = False

def chexdiff32(prev, cur, ascii=True, offset=0, offset2=None):
    assert len(cur) % 4 == 0
    count = len(cur) // 4
    words = struct.unpack("<%dI" % count, cur)

    if prev is None:
        last = None
    else:
        assert len(prev) == len(cur)
        last = struct.unpack("<%dI" % count, prev)

    row = 8
    skipping = False
    out = []
    for i in range(0, count, row):
        off_text = f"{offset + i * 4:016x}"
        if offset2 is not None:
            off_text += f"/{offset2 + i * 4:08x}"
        if not last:
            if i != 0 and words[i:i+row] == words[i-row:i]:
                if not skipping:
                    out.append(f"{off_text} *\n")
                skipping = True
            else:
                out.append(f"{off_text} ")
                for new in words[i:i+row]:
                    out.append("%08x " % new)
                if ascii:
                    out.append("| " + _ascii(cur[4*i:4*(i+row)]))
                out.append("\n")
                skipping = False
        elif last[i:i+row] != words[i:i+row]:
            out.append(f"{off_text} ")
            for old, new in zip(last[i:i+row], words[i:i+row]):
                so = "%08x" % old
                sn = s = "%08x" % new
                if old != new:
                    s = "\x1b[32m"
                    ld = False
                    for a,b in zip(so, sn):
                        d = a != b
                        if ld != d:
                            s += "\x1b[31;1;4m" if d else "\x1b[32m"
                            ld = d
                        s += b
                    s += "\x1b[m"
                out.append(s + " ")
            if ascii:
                out.append("| " + _ascii(cur[4*i:4*(i+row)]))
            out.append("\n")
    return "".join(out)

def chexundump(dump, base=0):
    if type(dump) is bytes:
        dump = dump.decode("ascii")
    elif type(dump) is str:
        pass
    else:
        dump = dump.read()

    decoded = bytearray()
    for line in dump.strip().splitlines():
        if not line:
            continue
        try:
            cropped = line.strip().split("|", 2)[0]
            mark, data = cropped.split(" ", 1)
            if data.strip() == "*":
                continue
            offset = int(mark, 16)
            data = data.replace(" ", "")
            if len(data) % 2 != 0:
                raise ValueError("odd sized data")
            if offset > len(decoded):
                decoded.extend([0] * (offset - len(decoded) - base))
            decoded.extend([int(data[i:i+2], 16) for i \
                            in range(0, len(data), 2)])
        except (ValueError, TypeError) as exc:
            raise ValueError(f"can't decode line: {line:r}") from exc

    return decoded

_extascii_table_low = [
    "▪", "☺", "☻", "♥", "♦", "♣", "♠", "•",
    "◘", "○", "◙", "♂", "♀", "♪", "♫", "☼",
    "►", "◄", "↕", "‼", "¶", "§", "▬", "↨",
    "↑", "↓", "→", "←", "∟", "↔", "▲", "▼"]

_extascii_table_high = [
    "⌂",
    "█", "⡀", "⢀", "⣀", "⠠", "⡠", "⢠", "⣠",
    "⠄", "⡄", "⢄", "⣄", "⠤", "⡤", "⢤", "⣤",
    "⠁", "⡁", "⢁", "⣁", "⠡", "⡡", "⢡", "⣡",
    "⠅", "⡅", "⢅", "⣅", "⠥", "⡥", "⢥", "⣥",
    "⠃", "⡃", "⢃", "⣃", "⠣", "⡣", "⢣", "⣣",
    "⠇", "⡇", "⢇", "⣇", "⠧", "⡧", "⢧", "⣧",
    "⠉", "⡉", "⢉", "⣉", "⠩", "⡩", "⢩", "⣩",
    "⠍", "⡍", "⢍", "⣍", "⠭", "⡭", "⢭", "⣭",
    "⠊", "⡊", "⢊", "⣊", "⠪", "⡪", "⢪", "⣪",
    "⠎", "⡎", "⢎", "⣎", "⠮", "⡮", "⢮", "⣮",
    "⠑", "⡑", "⢑", "⣑", "⠱", "⡱", "⢱", "⣱",
    "⠕", "⡕", "⢕", "⣕", "⠵", "⡵", "⢵", "⣵",
    "⠚", "⡚", "⢚", "⣚", "⠺", "⡺", "⢺", "⣺",
    "⠞", "⡞", "⢞", "⣞", "⠾", "⡾", "⢾", "⣾",
    "⠛", "⡛", "⢛", "⣛", "⠻", "⡻", "⢻", "⣻",
    "⠟", "⡟", "⢟", "⣟", "⠿", "⡿", "⢿", "⣿"]

def _extascii(s):
    s2 = ""
    for c in s:
        if c < 0x20:
            s2 += _extascii_table_low[c]
        elif c > 0x7e:
            s2 += _extascii_table_high[c-0x7f]
        else:
            s2 += chr(c)
    return s2

def ehexdump(s, st=0, abbreviate=True, indent="", print_fn=print):
    last = None
    skip = False
    for i in range(0,len(s),16):
        val = s[i:i+16]
        if val == last and abbreviate:
            if not skip:
                print_fn(indent+"%08x  *" % (i + st))
                skip = True
        else:
            print_fn(indent+"%08x  %s  %s  |%s|" % (
                  i + st,
                  hexdump(val[:8], ' ').ljust(23),
                  hexdump(val[8:], ' ').ljust(23),
                  _extascii(val).ljust(16)))
            last = val
            skip = False

def chexdump32(s, st=0, abbreviate=True, print_fn=print):
    last = None
    skip = False
    for i in range(0,len(s),32):
        val = s[i:i+32]
        if val == last and abbreviate:
            if not skip:
                print_fn("%08x  *" % (i + st))
                skip = True
        else:
            print_fn("%08x  %s" % (
                i + st,
                hexdump32(val, ' ')))
            last = val
            skip = False

def unhex(s):
    s = re.sub(r"/\*.*?\*/", "", s)
    return bytes.fromhex(s.replace(" ", "").replace("\n", ""))

def dumpstacks(signal, frame):
    id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
    code = []
    for threadId, stack in sys._current_frames().items():
        code.append("\n# Thread: %s(%d)" % (id2name.get(threadId,""), threadId))
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                code.append("  %s" % (line.strip()))
    print("\n".join(code))
    sys.exit(1)

def set_sigquit_stackdump_handler():
    signal.signal(signal.SIGQUIT, dumpstacks)

def parse_indexlist(s):
    items = set()
    for i in s.split(","):
        if "-" in i:
            a, b = map(int, i.split("-", 1))
            for i in range(a, b + 1):
                items.add(i)
        else:
            items.add(int(i))
    return items

FourCC = ExprAdapter(Int32ul,
                     lambda d, ctx: d.to_bytes(4, "big").decode("latin-1"),
                     lambda d, ctx: int.from_bytes(d.encode("latin-1"), "big"))

class SafeGreedyRange(GreedyRange):
    def __init__(self, subcon, discard=False):
        super().__init__(subcon)
        self.discard = discard

    def _parse(self, stream, context, path):
        discard = self.discard
        obj = ListContainer()
        try:
            for i in itertools.count():
                context._index = i
                e = self.subcon._parsereport(stream, context, path)
                if not discard:
                    obj.append(e)
        except StreamError:
            pass
        return obj

class ReloadableMeta(type):
    def __new__(cls, name, bases, dct):
        m = super().__new__(cls, name, bases, dct)
        m._load_time = time.time()
        return m

class Reloadable(metaclass=ReloadableMeta):
    @classmethod
    def _reloadcls(cls, force=False):
        mods = []
        for c in cls.mro():
            mod = sys.modules[c.__module__]
            cur_cls = getattr(mod, c.__name__)
            mods.append((cur_cls, mod))
            if c.__name__ == "Reloadable":
                break

        reloaded = set()
        newest = 0
        for pcls, mod in mods[::-1]:
            source = getattr(mod, "__file__", None)
            if not source:
                continue
            newest = max(newest, os.stat(source).st_mtime, pcls._load_time)
            if (force or reloaded or pcls._load_time < newest) and mod.__name__ not in reloaded:
                print(f"Reload: {mod.__name__}")
                mod = importlib.reload(mod)
                reloaded.add(mod.__name__)

        return getattr(mods[0][1], cls.__name__)

    def _reloadme(self):
        self.__class__ = self._reloadcls()

class Constant:
    def __init__(self, value):
        self.value = value

    def __call__(self, v):
        assert v == self.value
        return v

class RegisterMeta(ReloadableMeta):
    def __new__(cls, name, bases, dct):
        m = super().__new__(cls, name, bases, dct)

        f = {}

        if bases and bases[0] is not Reloadable:
            for cls in bases[0].mro():
                if cls is Reloadable:
                    break
                f.update({k: None for k,v in cls.__dict__.items()
                          if not k.startswith("_") and isinstance(v, (int, tuple))})

        f.update({k: None for k, v in dct.items()
                 if not k.startswith("_") and isinstance(v, (int, tuple))})

        m._fields_list = list(f.keys())
        m._fields = set(f.keys())

        return m

class Register(Reloadable, metaclass=RegisterMeta):
    _Constant = Constant
    def __init__(self, v=None, **kwargs):
        if v is not None:
            self._value = v
            for k in self._fields_list:
                getattr(self, k) # validate
        else:
            self._value = 0
            for k in self._fields_list:
                field = getattr(self.__class__, k)
                if isinstance(field, tuple) and len(field) >= 3 and isinstance(field[2], self._Constant):
                    setattr(self, k, field[2].value)

        for k,v in kwargs.items():
            setattr(self, k, v)

    def __getattribute__(self, attr):
        if attr.startswith("_") or attr not in self._fields:
            return object.__getattribute__(self, attr)

        field = getattr(self.__class__, attr)
        value = self._value

        if isinstance(field, int):
            return (value >> field) & 1
        elif isinstance(field, tuple):
            if len(field) == 2:
                msb, lsb = field
                ftype = int
            else:
                msb, lsb, ftype = field
            return ftype((value >> lsb) & ((1 << ((msb + 1) - lsb)) - 1))
        else:
            raise AttributeError(f"Invalid field definition {attr} = {field!r}")

    def __setattr__(self, attr, fvalue):
        if attr.startswith("_"):
            self.__dict__[attr] = fvalue
            return

        field = getattr(self.__class__, attr)

        value = self._value

        if isinstance(field, int):
            self._value = (value & ~(1 << field)) | ((fvalue & 1) << field)
        elif isinstance(field, tuple):
            if len(field) == 2:
                msb, lsb = field
            else:
                msb, lsb, ftype = field
            mask = ((1 << ((msb + 1) - lsb)) - 1)
            self._value = (value & ~(mask << lsb)) | ((fvalue & mask) << lsb)
        else:
            raise AttributeError(f"Invalid field definition {attr} = {field!r}")

    def __int__(self):
        return self._value

    def _field_val(self, field_name, as_repr=False):
        field = getattr(self.__class__, field_name)
        val = getattr(self, field_name)
        if isinstance(val, Enum):
            if as_repr:
                return str(val)
            else:
                msb, lsb = field[:2]
                if (msb - lsb + 1) > 3:
                    return f"0x{val.value:x}({val.name})"
                else:
                    return f"{val.value}({val.name})"
        elif not isinstance(val, int):
            return val
        elif isinstance(field, int):
            return val
        elif isinstance(field, tuple):
            msb, lsb = field[:2]
            if (msb - lsb + 1) > 3:
                return f"0x{val:x}"

        return val

    @property
    def fields(self):
        return {k: getattr(self, k) for k in self._fields_list}

    def str_fields(self):
        return ', '.join(f'{k}={self._field_val(k)}' for k in self._fields_list)

    def __str__(self):
        return f"0x{self._value:x} ({self.str_fields()})"

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(f'{k}={self._field_val(k, True)}' for k in self._fields_list)})"

    def copy(self):
        return type(self)(self._value)

    @property
    def value(self):
        return self._value
    @value.setter
    def value(self, val):
        self._value = val

class Register8(Register):
    __WIDTH__ = 8

class Register16(Register):
    __WIDTH__ = 16

class Register32(Register):
    __WIDTH__ = 32

class Register64(Register):
    __WIDTH__ = 64

class RegAdapter(Adapter):
    def __init__(self, register):
        if register.__WIDTH__ == 64:
            subcon = Int64ul
        elif register.__WIDTH__ == 32:
            subcon = Int32ul
        elif register.__WIDTH__ == 16:
            subcon = Int16ul
        elif register.__WIDTH__ == 8:
            subcon = Int8ul
        else:
            raise ValueError("Invalid reg width")

        self.reg = register
        super().__init__(subcon)

    def _decode(self, obj, context, path):
        return self.reg(obj)

    def _encode(self, obj, context, path):
        return obj.value

class RangeMap(Reloadable):
    def __init__(self):
        self.__start = []
        self.__end = []
        self.__value = []

    def clone(self):
        r = type(self)()
        r.__start = list(self.__start)
        r.__end = list(self.__end)
        r.__value = [copy.copy(i) for i in self.__value]
        return r

    def __len__(self):
        return len(self.__start)

    def __nonzero__(self):
        return bool(self.__start)

    def __contains(self, pos, addr):
        if pos < 0 or pos >= len(self.__start):
            return False

        return self.__start[pos] <= addr and addr <= self.__end[pos]

    def __split(self, pos, addr):
        self.__start.insert(pos + 1, addr)
        self.__end.insert(pos, addr - 1)
        self.__value.insert(pos + 1, copy.copy(self.__value[pos]))

    def __zone(self, zone):
        if isinstance(zone, slice):
            zone = range(zone.start if zone.start is not None else 0,
                         zone.stop if zone.stop is not None else 1 << 64)
        elif isinstance(zone, int):
            zone = range(zone, zone + 1)

        return zone

    def lookup(self, addr, default=None):
        addr = int(addr)

        pos = bisect.bisect_left(self.__end, addr)
        if self.__contains(pos, addr):
            return self.__value[pos]
        else:
            return default

    def __iter__(self):
        return self.ranges()

    def ranges(self):
        return (range(s, e + 1) for s, e in zip(self.__start, self.__end))

    def items(self):
        return ((range(s, e + 1), v) for s, e, v in zip(self.__start, self.__end, self.__value))

    def _overlap_range(self, zone, split=False):
        zone = self.__zone(zone)
        if not zone:
            return 0, 0

        start = bisect.bisect_left(self.__end, zone.start)

        if split:
            # Handle left-side overlap
            if self.__contains(start, zone.start) and self.__start[start] != zone.start:
                self.__split(start, zone.start)
                start += 1
                assert self.__start[start] == zone.start

        for pos in range(start, len(self.__start)):
            if self.__start[pos] >= zone.stop:
                return start, pos
            if split and (self.__end[pos] + 1) > zone.stop:
                self.__split(pos, zone.stop)
                return start, pos + 1

        return start, len(self.__start)

    def populate(self, zone, default=[]):
        zone = self.__zone(zone)
        if len(zone) == 0:
            return

        start, stop = zone.start, zone.stop

        # Starting insertion point, overlap inclusive
        pos = bisect.bisect_left(self.__end, zone.start)

        # Handle left-side overlap
        if self.__contains(pos, zone.start) and self.__start[pos] != zone.start:
            self.__split(pos, zone.start)
            pos += 1
            assert self.__start[pos] == zone.start

        # Iterate through overlapping ranges
        while start < stop:
            if pos == len(self.__start):
                # Append to end
                val = copy.copy(default)
                self.__start.append(start)
                self.__end.append(stop - 1)
                self.__value.append(val)
                yield range(start, stop), val
                break

            assert self.__start[pos] >= start
            if self.__start[pos] > start:
                # Insert new range
                boundary = stop
                if pos < len(self.__start):
                    boundary = min(stop, self.__start[pos])
                val = copy.copy(default)
                self.__start.insert(pos, start)
                self.__end.insert(pos, boundary - 1)
                self.__value.insert(pos, val)
                yield range(start, boundary), val
                start = boundary
            else:
                # Handle right-side overlap
                if self.__end[pos] > stop - 1:
                    self.__split(pos, stop)
                # Add to existing range
                yield range(self.__start[pos], self.__end[pos] + 1), self.__value[pos]
                start = self.__end[pos] + 1

            pos += 1
        else:
            assert start == stop

    def overlaps(self, zone, split=False):
        start, stop = self._overlap_range(zone, split)
        for pos in range(start, stop):
            yield range(self.__start[pos], self.__end[pos] + 1), self.__value[pos]

    def replace(self, zone, val):
        zone = self.__zone(zone)
        if zone.start == zone.stop:
            return
        start, stop = self._overlap_range(zone, True)
        self.__start = self.__start[:start] + [zone.start] + self.__start[stop:]
        self.__end = self.__end[:start] + [zone.stop - 1] + self.__end[stop:]
        self.__value = self.__value[:start] + [val] + self.__value[stop:]

    def clear(self, zone=None):
        if zone is None:
            self.__start = []
            self.__end = []
            self.__value = []
        else:
            zone = self.__zone(zone)
            if zone.start == zone.stop:
                return
            start, stop = self._overlap_range(zone, True)
            self.__start = self.__start[:start] + self.__start[stop:]
            self.__end = self.__end[:start] + self.__end[stop:]
            self.__value = self.__value[:start] + self.__value[stop:]

    def compact(self, equal=lambda a, b: a == b, empty=lambda a: not a):
        if len(self) == 0:
            return

        new_s, new_e, new_v = [], [], []

        for pos in range(len(self)):
            s, e, v = self.__start[pos], self.__end[pos], self.__value[pos]
            if empty(v):
                continue
            if new_v and equal(last, v) and s == new_e[-1] + 1:
                new_e[-1] = e
            else:
                new_s.append(s)
                new_e.append(e)
                new_v.append(v)
                last = v

        self.__start, self.__end, self.__value = new_s, new_e, new_v

    def _assert(self, expect, val=lambda a:a):
        state = []
        for i, j, v in zip(self.__start, self.__end, self.__value):
            state.append((i, j, val(v)))
        if state != expect:
            print(f"Expected: {expect}")
            print(f"Got:      {state}")

class AddrLookup(RangeMap):
    def __str__(self):
        b = [""]
        for zone, values in self.items():
            b.append(f"{zone.start:#11x} - {zone.stop - 1:#11x}")
            if len(values) == 0:
                b.append(f" (empty range)")
            elif len(values) == 1:
                b.append(f" : {values[0][0]}\n")
            if len(values) > 1:
                b.append(f" ({len(values):d} devices)\n")
                for value, r in sorted(values, key=lambda r: r[1].start):
                    b.append(f"      {r.start:#10x} - {r.stop - 1:#8x} : {value}\n")

        return "".join(b)

    def add(self, zone, value):
        for r, values in self.populate(zone):
            values.append((value, zone))

    def remove(self, zone, value):
        for r, values in self.overlaps(zone):
            try:
                values.remove((value, zone))
            except:
                pass

    def lookup(self, addr, default='unknown'):
        maps = super().lookup(addr)
        return maps[0] if maps else (default, range(0, 1 << 64))

    def lookup_all(self, addr):
        return super().lookup(addr, [])

    def _assert(self, expect, val=lambda a:a):
        super()._assert(expect, lambda v: [i[0] for i in v])

class ScalarRangeMap(RangeMap):
    def get(self, addr, default=None):
        return self.lookup(addr, default)

    def __setitem__(self, zone, value):
        self.replace(zone, value)

    def __delitem__(self, zone):
        self.clear(zone)

    def __getitem__(self, addr):
        value = self.lookup(addr, default=KeyError)
        if value is KeyError:
            raise KeyError(f"Address {addr:#x} has no value")
        return value

class BoolRangeMap(RangeMap):
    def set(self, zone):
        self.replace(zone, True)

    def __delitem__(self, zone):
        self.clear(zone)

    def __getitem__(self, addr):
        return self.lookup(addr, False)

class DictRangeMap(RangeMap):
    def __setitem__(self, k, value):
        if not isinstance(k, tuple):
            self.replace(k, dict(value))
        else:
            zone, key = k
            for r, values in self.populate(zone, {}):
                values[key] = value

    def __delitem__(self, k):
        if not isinstance(k, tuple):
            self.clear(k)
        else:
            zone, key = k
            for r, values in self.overlaps(zone, True):
                values.pop(key, None)

    def __getitem__(self, k):
        if isinstance(k, tuple):
            addr, k = k
            values = self.lookup(addr)
            return values.get(k, None) if values else None
        else:
            values = self.lookup(k)
            return values or {}

class SetRangeMap(RangeMap):
    def add(self, zone, key):
        for r, values in self.populate(zone, set()):
            values.add(key)

    def discard(self, zone, key):
        for r, values in self.overlaps(zone, split=True):
            if values:
                values.discard(key)
    remove = discard

    def __setitem__(self, k, value):
        self.replace(k, set(value))

    def __delitem__(self, k):
        self.clear(k)

    def __getitem__(self, addr):
        values = super().lookup(addr)
        return frozenset(values) if values else frozenset()

class NdRange:
    def __init__(self, rng, min_step=1):
        if isinstance(rng, range):
            self.ranges = [rng]
        else:
            self.ranges = list(rng)
        least_step = self.ranges[0].step
        for i, rng in enumerate(self.ranges):
            if rng.step == 1:
                self.ranges[i] = range(rng.start, rng.stop, min_step)
                least_step = min_step
            else:
                assert rng.step >= min_step
                least_step = min(least_step, rng.step)
        self.start = sum(rng[0] for rng in self.ranges)
        self.stop = sum(rng[-1] for rng in self.ranges) + least_step
        self.rev = {}
        for i in itertools.product(*map(enumerate, self.ranges)):
            index = tuple(j[0] for j in i)
            addr = sum(j[1] for j in i)
            if len(self.ranges) == 1:
                index = index[0]
            self.rev[addr] = index

    def index(self, item):
        return self.rev[item]

    def __len__(self):
        return self.stop - self.start

    def __contains__(self, item):
        return item in self.rev

    def __getitem__(self, item):
        if not isinstance(item, tuple):
            assert len(self.ranges) == 1
            return self.ranges[0][item]

        assert len(self.ranges) == len(item)
        if all(isinstance(i, int) for i in item):
            return sum((i[j] for i, j in zip(self.ranges, item)))
        else:
            iters = (i[j] for i, j in zip(self.ranges, item))
            return map(sum, itertools.product(*(([i] if isinstance(i, int) else i) for i in iters)))

class RegMapMeta(ReloadableMeta):
    def __new__(cls, name, bases, dct):
        m = super().__new__(cls, name, bases, dct)
        if getattr(m, "_addrmap", None) is None:
            m._addrmap = {}
            m._rngmap = SetRangeMap()
            m._namemap = {}
        else:
            m._addrmap = dict(m._addrmap)
            m._rngmap = m._rngmap.clone()
            m._namemap = dict(m._namemap)

        for k, v in dct.items():
            if k.startswith("_") or not isinstance(v, tuple):
                continue
            addr, rtype = v

            if isinstance(addr, int):
                m._addrmap[addr] = k, rtype
            else:
                addr = NdRange(addr, rtype.__WIDTH__ // 8)
                m._rngmap.add(addr, (addr, k, rtype))

            m._namemap[k] = addr, rtype

            def prop(k):
                def getter(self):
                    return self._accessor[k]
                def setter(self, val):
                    self._accessor[k].val = val
                return property(getter, setter)

            setattr(m, k, prop(k))

        return m

class RegAccessor(Reloadable):
    def __init__(self, cls, rd, wr, addr):
        self.cls = cls
        self.rd = rd
        self.wr = wr
        self.addr = addr

    def __int__(self):
        return self.rd(self.addr)

    @property
    def val(self):
        return self.rd(self.addr)

    @val.setter
    def val(self, value):
        self.wr(self.addr, int(value))

    @property
    def reg(self):
        val = self.val
        if val is None:
            return None
        return self.cls(val)

    @reg.setter
    def reg(self, value):
        self.wr(self.addr, int(value))

    def set(self, **kwargs):
        r = self.reg
        for k, v in kwargs.items():
            setattr(r, k, v)
        self.wr(self.addr, int(r))

    def __str__(self):
        return str(self.reg)

class RegArrayAccessor(Reloadable):
    def __init__(self, range, cls, rd, wr, addr):
        self.range = range
        self.cls = cls
        self.rd = rd
        self.wr = wr
        self.addr = addr

    def __getitem__(self, item):
        off = self.range[item]
        if isinstance(off, int):
            return RegAccessor(self.cls, self.rd, self.wr, self.addr + off)
        else:
            return [RegAccessor(self.cls, self.rd, self.wr, self.addr + i) for i in off]

class BaseRegMap(Reloadable):
    def __init__(self, backend, base):
        self._base = base
        self._backend = backend
        self._accessor = {}

        for name, (addr, rcls) in self._namemap.items():
            width = rcls.__WIDTH__
            rd = functools.partial(backend.read, width=width)
            wr = functools.partial(backend.write, width=width)
            if type(addr).__name__ == "NdRange":
                self._accessor[name] = RegArrayAccessor(addr, rcls, rd, wr, base)
            else:
                self._accessor[name] = RegAccessor(rcls, rd, wr, base + addr)

    def _lookup_offset(cls, offset):
        reg = cls._addrmap.get(offset, None)
        if reg is not None:
            name, rcls = reg
            return name, None, rcls
        ret = cls._rngmap[offset]
        if ret:
            for rng, name, rcls in ret:
                if offset in rng:
                    return name, rng.index(offset), rcls
        return None, None, None
    lookup_offset = classmethod(_lookup_offset)

    def lookup_addr(self, addr):
        return self.lookup_offset(addr - self._base)

    def get_name(self, addr):
        name, index, rcls = self.lookup_addr(addr)
        if index is not None:
            return f"{name}[{index}]"
        else:
            return name

    def _lookup_name(cls, name):
        return cls._namemap.get(name, None)
    lookup_name = classmethod(_lookup_name)

    def _scalar_regs(self):
        for addr, (name, rtype) in self._addrmap.items():
            yield addr, name, self._accessor[name], rtype

    def _array_reg(self, zone, map):
        addrs, name, rtype = map
        def index(addr):
            idx = addrs.index(addr)
            if isinstance(idx, tuple):
                idx = str(idx)[1:-1]
            return idx
        reg = ((addr, f"{name}[{index(addr)}]", self._accessor[name][addrs.index(addr)], rtype)
                     for addr in zone if addr in addrs)
        return reg

    def _array_regs(self):
        for zone, maps in self._rngmap.items():
            yield from heapq.merge(*(self._array_reg(zone, map) for map in maps))

    def dump_regs(self):
        for addr, name, acc, rtype in heapq.merge(sorted(self._scalar_regs()), self._array_regs()):
            print(f"{self._base:#x}+{addr:06x} {name} = {acc.reg}")

class RegMap(BaseRegMap, metaclass=RegMapMeta):
    pass

def irange(start, count, step=1):
    return range(start, start + count * step, step)

# Table generated by:
#
# tbl = [0] * 256
# crc = 1
# for i in [2**x for x in irange(7, 0, -1)]:
#    if crc & 1:
#         crc = (crc >> 1) ^ 0xA001
#     else:
#         crc = crc >> 1
#     for j in range(0, 255, 2*i):
#         tbl[i + j] = crc ^ tbl[j]
#
# for i in range(0, 255, 8):
#     print(f"{tbl[i]:#06x}, {tbl[i+1]:#06x}, {tbl[i+2]:#06x}, {tbl[i+3]:#06x}, {tbl[i+4]:#06x}, {tbl[i+5]:#06x}, {tbl[i+6]:#06x}, {tbl[i+7]:#06x}, ")

_crc16_table = [
    0x0000, 0xc0c1, 0xc181, 0x0140, 0xc301, 0x03c0, 0x0280, 0xc241,
    0xc601, 0x06c0, 0x0780, 0xc741, 0x0500, 0xc5c1, 0xc481, 0x0440,
    0xcc01, 0x0cc0, 0x0d80, 0xcd41, 0x0f00, 0xcfc1, 0xce81, 0x0e40,
    0x0a00, 0xcac1, 0xcb81, 0x0b40, 0xc901, 0x09c0, 0x0880, 0xc841,
    0xd801, 0x18c0, 0x1980, 0xd941, 0x1b00, 0xdbc1, 0xda81, 0x1a40,
    0x1e00, 0xdec1, 0xdf81, 0x1f40, 0xdd01, 0x1dc0, 0x1c80, 0xdc41,
    0x1400, 0xd4c1, 0xd581, 0x1540, 0xd701, 0x17c0, 0x1680, 0xd641,
    0xd201, 0x12c0, 0x1380, 0xd341, 0x1100, 0xd1c1, 0xd081, 0x1040,
    0xf001, 0x30c0, 0x3180, 0xf141, 0x3300, 0xf3c1, 0xf281, 0x3240,
    0x3600, 0xf6c1, 0xf781, 0x3740, 0xf501, 0x35c0, 0x3480, 0xf441,
    0x3c00, 0xfcc1, 0xfd81, 0x3d40, 0xff01, 0x3fc0, 0x3e80, 0xfe41,
    0xfa01, 0x3ac0, 0x3b80, 0xfb41, 0x3900, 0xf9c1, 0xf881, 0x3840,
    0x2800, 0xe8c1, 0xe981, 0x2940, 0xeb01, 0x2bc0, 0x2a80, 0xea41,
    0xee01, 0x2ec0, 0x2f80, 0xef41, 0x2d00, 0xedc1, 0xec81, 0x2c40,
    0xe401, 0x24c0, 0x2580, 0xe541, 0x2700, 0xe7c1, 0xe681, 0x2640,
    0x2200, 0xe2c1, 0xe381, 0x2340, 0xe101, 0x21c0, 0x2080, 0xe041,
    0xa001, 0x60c0, 0x6180, 0xa141, 0x6300, 0xa3c1, 0xa281, 0x6240,
    0x6600, 0xa6c1, 0xa781, 0x6740, 0xa501, 0x65c0, 0x6480, 0xa441,
    0x6c00, 0xacc1, 0xad81, 0x6d40, 0xaf01, 0x6fc0, 0x6e80, 0xae41,
    0xaa01, 0x6ac0, 0x6b80, 0xab41, 0x6900, 0xa9c1, 0xa881, 0x6840,
    0x7800, 0xb8c1, 0xb981, 0x7940, 0xbb01, 0x7bc0, 0x7a80, 0xba41,
    0xbe01, 0x7ec0, 0x7f80, 0xbf41, 0x7d00, 0xbdc1, 0xbc81, 0x7c40,
    0xb401, 0x74c0, 0x7580, 0xb541, 0x7700, 0xb7c1, 0xb681, 0x7640,
    0x7200, 0xb2c1, 0xb381, 0x7340, 0xb101, 0x71c0, 0x7080, 0xb041,
    0x5000, 0x90c1, 0x9181, 0x5140, 0x9301, 0x53c0, 0x5280, 0x9241,
    0x9601, 0x56c0, 0x5780, 0x9741, 0x5500, 0x95c1, 0x9481, 0x5440,
    0x9c01, 0x5cc0, 0x5d80, 0x9d41, 0x5f00, 0x9fc1, 0x9e81, 0x5e40,
    0x5a00, 0x9ac1, 0x9b81, 0x5b40, 0x9901, 0x59c0, 0x5880, 0x9841,
    0x8801, 0x48c0, 0x4980, 0x8941, 0x4b00, 0x8bc1, 0x8a81, 0x4a40,
    0x4e00, 0x8ec1, 0x8f81, 0x4f40, 0x8d01, 0x4dc0, 0x4c80, 0x8c41,
    0x4400, 0x84c1, 0x8581, 0x4540, 0x8701, 0x47c0, 0x4680, 0x8641,
    0x8201, 0x42c0, 0x4380, 0x8341, 0x4100, 0x81c1, 0x8081, 0x4040
]

def crc16USB(crc, data):
    for x in data:
        crc = (crc >> 8) ^ _crc16_table[(crc ^ x) & 0xff]
    return crc

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)

if __name__ == "__main__":
    # AddrLookup test
    a = AddrLookup()
    a.add(range(0, 10), 0)
    a._assert([
        (0, 9, [0])
    ])
    a.add(range(10, 20), 1)
    a._assert([
        (0, 9, [0]), (10, 19, [1])
    ])
    a.add(range(20, 25), 2)
    a._assert([
        (0, 9, [0]), (10, 19, [1]), (20, 24, [2])
    ])
    a.add(range(30, 40), 3)
    a._assert([
        (0, 9, [0]), (10, 19, [1]), (20, 24, [2]), (30, 39, [3])
    ])
    a.add(range(0, 15), 4)
    a._assert([
        (0, 9, [0, 4]), (10, 14, [1, 4]), (15, 19, [1]), (20, 24, [2]), (30, 39, [3])
    ])
    a.add(range(0, 15), 5)
    a._assert([
        (0, 9, [0, 4, 5]), (10, 14, [1, 4, 5]), (15, 19, [1]), (20, 24, [2]), (30, 39, [3])
    ])
    a.add(range(21, 44), 6)
    a._assert([
        (0, 9, [0, 4, 5]), (10, 14, [1, 4, 5]), (15, 19, [1]), (20, 20, [2]), (21, 24, [2, 6]),
        (25, 29, [6]), (30, 39, [3, 6]), (40, 43, [6])
    ])
    a.add(range(70, 80), 7)
    a._assert([
        (0, 9, [0, 4, 5]), (10, 14, [1, 4, 5]), (15, 19, [1]), (20, 20, [2]), (21, 24, [2, 6]),
        (25, 29, [6]), (30, 39, [3, 6]), (40, 43, [6]), (70, 79, [7])
    ])
    a.add(range(0, 100), 8)
    a._assert([
        (0, 9, [0, 4, 5, 8]), (10, 14, [1, 4, 5, 8]), (15, 19, [1, 8]), (20, 20, [2, 8]),
        (21, 24, [2, 6, 8]), (25, 29, [6, 8]), (30, 39, [3, 6, 8]), (40, 43, [6, 8]),
        (44, 69, [8]), (70, 79, [7, 8]), (80, 99, [8])
    ])
    a.remove(range(21, 44), 6)
    a._assert([
        (0, 9, [0, 4, 5, 8]), (10, 14, [1, 4, 5, 8]), (15, 19, [1, 8]), (20, 20, [2, 8]),
        (21, 24, [2, 8]), (25, 29, [8]), (30, 39, [3, 8]), (40, 43, [8]),
        (44, 69, [8]), (70, 79, [7, 8]), (80, 99, [8])
    ])
    a.compact()
    a._assert([
        (0, 9, [0, 4, 5, 8]), (10, 14, [1, 4, 5, 8]), (15, 19, [1, 8]), (20, 24, [2, 8]),
        (25, 29, [8]), (30, 39, [3, 8]), (40, 69, [8]), (70, 79, [7, 8]),
        (80, 99, [8])
    ])
    a.remove(range(0, 100), 8)
    a._assert([
        (0, 9, [0, 4, 5]), (10, 14, [1, 4, 5]), (15, 19, [1]), (20, 24, [2]), (25, 29, []),
        (30, 39, [3]), (40, 69, []), (70, 79, [7]), (80, 99, [])
    ])
    a.compact()
    a._assert([
        (0, 9, [0, 4, 5]), (10, 14, [1, 4, 5]), (15, 19, [1]), (20, 24, [2]), (30, 39, [3]),
        (70, 79, [7])
    ])
    a.clear(range(12, 21))
    a._assert([
        (0, 9, [0, 4, 5]), (10, 11, [1, 4, 5]), (21, 24, [2]), (30, 39, [3]),
        (70, 79, [7])
    ])

    # ScalarRangeMap test
    a = ScalarRangeMap()
    a[0:5] = 1
    a[5:10] = 2
    a[4:8] = 3
    del a[2:4]
    expect = [1, 1, None, None, 3, 3, 3, 3, 2, 2, None]
    for i,j in enumerate(expect):
        assert a.get(i) == j
        if j is not None:
            assert a[i] == j
    try:
        a[10]
    except KeyError:
        pass
    else:
        assert False

    # DictRangeMap test
    a = DictRangeMap()
    a[0:5, 0] = 10
    a[5:8, 1] = 11
    a[4:6, 2] = 12
    del a[2:4]
    expect = [{0: 10}, {0: 10}, {}, {}, {0: 10, 2: 12}, {1: 11, 2: 12}, {1: 11}, {1: 11}, {}]
    for i,j in enumerate(expect):
        assert a[i] == j
        for k, v in j.items():
            assert a[i, k] == v

    # SetRangeMap test
    a = SetRangeMap()
    a[0:2] = {1,}
    a[2:7] = {2,}
    a.add(range(1, 4), 3)
    a.discard(0, -1)
    a.discard(3, 2)
    del a[4]
    expect = [{1,}, {1,3}, {2,3}, {3,}, set(), {2,}, {2,}, set()]
    for i,j in enumerate(expect):
        assert a[i] == j

    # BoolRangeMap test
    a = BoolRangeMap()
    a.set(range(0, 2))
    a.set(range(4, 6))
    a.set(range(5, 5))
    a.clear(range(3, 5))
    expect = [True, True, False, False, False, True, False]
    for i,j in enumerate(expect):
        assert a[i] == j



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import serial
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import argparse, pathlib

from m1n1 import adt

parser = argparse.ArgumentParser(description='Convert ADT PMGR nodes to Device Tree format')
parser.add_argument("-m", "--multidie", action="store_true")
parser.add_argument('input', type=pathlib.Path)
args = parser.parse_args()

adt_data = args.input.read_bytes()
dt = adt.load_adt(adt_data)

pmgr = dt["/arm-io/pmgr"]

dev_by_id = {dev.id: dev for dev in pmgr.devices}

blocks = {}
maxaddr = {}

def die_node(s):
    if args.multidie:
        return f"DIE_NODE({s})"
    else:
        return s

def die_label(s):
    if args.multidie:
        return f"DIE_LABEL({s})"
    else:
        return s

for i, dev in enumerate(pmgr.devices):
    if dev.flags.no_ps:
        continue
    ps = pmgr.ps_regs[dev.psreg]
    block = pmgr.get_reg(ps.reg)
    blocks.setdefault(block, []).append(dev)
    offset = ps.offset + dev.psidx * 8
    maxaddr[block[0]] = max(maxaddr.get(block[0], 0), offset)

pmgr_compat = pmgr.compatible[0].split(",")[1]
compatible = f'"apple,{pmgr_compat}-pmgr", "apple,pmgr", "syscon", "simple-mfd"'
ps_compatible = f'"apple,{pmgr_compat}-pmgr-pwrstate", "apple,pmgr-pwrstate"'

for i, ((base, size), devices) in enumerate(sorted(blocks.items())):

    size = min(size, (maxaddr[base] + 0x3fff) & ~0x3fff)

    print(f"pmgr{i}: power-management@{base:x} {{")
    print(f"\tcompatible = {compatible};")
    print( "\t#address-cells = <1>;")
    print( "\t#size-cells = <1>;")
    print()
    print(f"\treg = <{base >> 32:#x} {base & 0xffffffff:#x} 0 {size:#x}>;")
    print( "};")
    print()

for i, ((base, size), devices) in enumerate(sorted(blocks.items())):
    print(f"&pmgr{i} {{")

    for dev in sorted(devices, key=lambda d: pmgr.ps_regs[d.psreg].offset + dev.psidx * 8):
        if dev.flags.no_ps:
            continue

        ps = pmgr.ps_regs[dev.psreg]
        offset = ps.offset + dev.psidx * 8
        addr = pmgr.get_reg(ps.reg)[0] + offset
        assert base <= addr <= (base + size)

        print()
        print(f"\t{die_node('ps_' + dev.name.lower())}: power-controller@{offset:x} {{")
        print(f"\t\tcompatible = {ps_compatible};")
        print(f"\t\treg = <{offset:#x} 4>;")
        print( "\t\t#power-domain-cells = <0>;")
        print( "\t\t#reset-cells = <0>;")
        print(f'\t\tlabel = {die_label(dev.name.lower())};')
        if dev.flags.critical:
            print("\t\tapple,always-on;")

        if any(dev.parents):
            domains = [f"<&{die_node('ps_'+dev_by_id[idx].name.lower())}>" for idx in dev.parents if idx]
            print(f"\t\tpower-domains = {', '.join(domains)};")

        print( "\t};")
    print( "};")
    print()



# SPDX-License-Identifier: MIT
import sys, pathlib

sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.hw.dart import DART
from m1n1.hw.aes import *

def aes_set_custom_key(
    aes,
    key,
    encrypt=True,
    mode=AES_SET_KEY_BLOCK_MODE.CTR,
    keyslot=0,
    keygen=0,
):
    keylen = {
        16: AES_SET_KEY_LEN.AES128,
        24: AES_SET_KEY_LEN.AES192,
        32: AES_SET_KEY_LEN.AES256,
    }[len(key)]

    aes.R_CMD_FIFO = AESSetKeyCommand(
        KEY_SELECT=0,
        KEYLEN=keylen,
        ENCRYPT=1 if encrypt else 0,
        BLOCK_MODE=mode,
        SLOT=keyslot,
        KEYGEN=keygen,
    ).value

    for i in range(0, len(key), 4):
        aes.R_CMD_FIFO = struct.unpack(">I", key[i : i + 4])[0]


def aes_set_hw_key(
    aes,
    key,
    keylen=AES_SET_KEY_LEN.AES128,
    encrypt=True,
    mode=AES_SET_KEY_BLOCK_MODE.CTR,
    slot=0,
    keygen=0,
):
    aes.R_CMD_FIFO = AESSetKeyCommand(
        KEY_SELECT=key,
        KEYLEN=keylen,
        ENCRYPT=1 if encrypt else 0,
        BLOCK_MODE=mode,
        SLOT=slot,
        KEYGEN=keygen,
    ).value


def aes_set_iv(aes, iv, slot=0):
    assert len(iv) == 16
    aes.R_CMD_FIFO = AESSetIVCommand(SLOT=slot)

    for i in range(0, len(iv), 4):
        aes.R_CMD_FIFO = struct.unpack(">I", iv[i : i + 4])[0]


def aes_crypt(aes, dart, data, key_slot=0, iv_slot=0):
    assert len(data) % 16 == 0

    bfr = p.memalign(0x4000, len(data))
    iova = dart.iomap(1, bfr, len(data))
    dart.iowrite(1, iova, data)

    aes.R_CMD_FIFO = AESCryptCommand(LEN=len(data), KEY_SLOT=key_slot, IV_SLOT=iv_slot)
    aes.R_CMD_FIFO = 0  # actually upper bits of addr
    aes.R_CMD_FIFO = iova  # src
    aes.R_CMD_FIFO = iova  # dst

    aes.R_CMD_FIFO = AESBarrierCommand(IRQ=1).value
    time.sleep(0.1)
    # while aes.R_IRQ_STATUS.reg.FLAG != 1:
    #    pass
    # aes.dump_regs()
    aes.R_IRQ_STATUS = aes.R_IRQ_STATUS.val

    res = dart.ioread(1, iova, len(data))
    return res


def test_hw_key(key, keylen, keygen=0):
    aes.R_IRQ_STATUS = aes.R_IRQ_STATUS.val
    aes.R_CONTROL.set(CLEAR_FIFO=1)
    aes.R_CONTROL.set(RESET=1)
    aes.R_CONTROL.set(START=1)
    # aes.dump_regs()
    aes_set_hw_key(aes, key, keylen, slot=0, keygen=keygen)
    # print(aes.R_IRQ_STATUS)
    aes_set_iv(aes, b"\x00" * 16, slot=0)
    chexdump(aes_crypt(aes, dart, b"\x00" * 16, key_slot=0, iv_slot=1))
    # aes.dump_regs()
    aes.R_CONTROL.set(STOP=1)


def test_custom_key(key, keygen=0):
    aes.R_IRQ_STATUS = aes.R_IRQ_STATUS.val
    aes.R_CONTROL.set(CLEAR_FIFO=1)
    aes.R_CONTROL.set(RESET=1)
    aes.R_CONTROL.set(START=1)
    # aes.dump_regs()
    aes_set_custom_key(aes, key, keyslot=0, keygen=keygen)
    aes_set_iv(aes, b"\x00" * 16)
    aes_set_iv(aes, b"\x11" * 16, slot=1)
    chexdump(aes_crypt(aes, dart, b"\x00" * 16, key_slot=0, iv_slot=0))
    # aes.dump_regs()
    aes.R_CONTROL.set(STOP=1)


p.pmgr_adt_clocks_enable("/arm-io/aes")

dart = DART.from_adt(u, "/arm-io/dart-sio")
dart.initialize()

aes_base, _ = u.adt["/arm-io/aes"].get_reg(0)
aes = AESRegs(u, aes_base)
aes.dump_regs()

dart.dump_all()

for keygen in range(4):
    print(f"zero key, keygen={keygen}", end="")
    test_custom_key(b"\x00" * 16, keygen=keygen)

for keygen in range(4):
    print("#" * 10)
    for keylen in [
        AES_SET_KEY_LEN.AES128,
        AES_SET_KEY_LEN.AES192,
        AES_SET_KEY_LEN.AES256,
    ]:
        for i in (1, 3):
            print(f"key = {i}, keylen={keylen}, keygen={keygen}", end="")
            test_hw_key(i, keylen, keygen=keygen)

dart.dump_all()

run_shell(globals(), msg="Have fun!")



# SPDX-License-Identifier: MIT

from m1n1.utils import irange

# Map the entire MMIO range as traceable
for r in hv.adt["/arm-io"].ranges:
    trace_range(irange(r.parent_addr, r.size), mode=TraceMode.ASYNC)

# Skip some noisy devices
try:
    trace_device("/arm-io/usb-drd0", False)
except KeyError:
    pass
try:
    trace_device("/arm-io/usb-drd1", False)
except KeyError:
    pass
try:
    trace_device("/arm-io/uart2", False)
except KeyError:
    pass
trace_device("/arm-io/error-handler", False)
trace_device("/arm-io/aic", False)
trace_device("/arm-io/spi1", False)
trace_device("/arm-io/pmgr", False)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
from m1n1.trace.isp import ISPTracer

# bootargs += camLogging=0xfffffffffffff
# all opcodes: https://pastebin.com/F8h7eDRB
dev_name = [node.name for node in u.adt["/arm-io"] if node.name.startswith("isp")][0]
dev_path = "/arm-io/%s" % (dev_name)
dart_dev_path = "/arm-io/dart-%s" % (dev_name)
print(dev_path, dart_dev_path)
tracer = ISPTracer(hv, dev_path, dart_dev_path, verbose=4)
tracer.start()
dart = tracer.dart_tracer.dart



# SPDX-License-Identifier: MIT
from ...utils import *
from ...malloc import Heap

from ..asc import StandardASC
from ..asc.base import ASCBaseEndpoint, msg_handler

from .initdata import InitData, IOMapping

__all__ = []

class GpuMsg(Register64):
    TYPE    = 63, 48

class InitMsg(GpuMsg):
    TYPE        = 63, 48, Constant(0x81)
    UNK         = 47, 44, Constant(0)
    INITDATA    = 43, 0

class EventMsg(GpuMsg):
    TYPE        = 63, 48, Constant(0x42)
    UNK         = 47, 0, Constant(0)

class DoorbellMsg(GpuMsg):
    TYPE        = 63, 48, Constant(0x83)
    CHANNEL     = 15, 0

class FWCtlMsg(GpuMsg):
    TYPE        = 63, 48, Constant(0x84)

class HaltMsg(GpuMsg):
    TYPE        = 63, 48, Constant(0x85)

class FirmwareEP(ASCBaseEndpoint):
    BASE_MESSAGE = GpuMsg
    SHORT = "fw"

    @msg_handler(0x42)
    def event(self, msg):
        #self.log("Received event")
        self.asc.agx.poll_channels()
        return True

    def send_initdata(self, addr):
        self.log(f"Sending initdata @ {addr:#x}")
        msg = InitMsg(INITDATA=addr)
        self.send(msg)

class DoorbellEP(ASCBaseEndpoint):
    BASE_MESSAGE = DoorbellMsg
    SHORT = "db"

    def doorbell(self, channel):
        #self.log(f"Sending doorbell ch={channel}")
        msg = DoorbellMsg(CHANNEL = channel)
        self.send(msg)

    def fwctl_doorbell(self):
        msg = FWCtlMsg()
        self.send(msg)

class AGXASC(StandardASC):
    ENDPOINTS = {
        0x20: FirmwareEP,
        0x21: DoorbellEP,
    }

    def __init__(self, u, base, agx, uat):
        super().__init__(u, base)
        self.agx = agx
        self.uat = uat

    def addr(self, addr):
        base, obj = self.agx.find_object(addr)
        if base is None:
            return super().addr(addr)

        return f"{addr:#x} ({obj._name} [{obj._size:#x}] @ {base:#x} + {addr - base:#x})"

    def iomap(self, addr, size):
        return self.uat.iomap(0, addr, size)

    def ioalloc(self, size):
        paddr = self.u.memalign(0x4000, size)
        dva = self.iomap(paddr, size)
        return paddr, dva

    def ioread(self, dva, size, ctx=0):
        return self.uat.ioread(ctx, dva & 0xFFFFFFFFFF, size)

    def iowrite(self, dva, data, ctx=0):
        return self.uat.iowrite(ctx, dva & 0xFFFFFFFFFF, data)

    def iotranslate(self, dva, size, ctx=0):
        return self.uat.iotranslate(ctx, dva & 0xFFFFFFFFFF, size)

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)
